﻿#light

open System
open System.Collections
open System.Collections.Generic;
open System.Text;
open Microsoft.FSharp.Math;
open System.Collections.Generic
open System.IO
open Microsoft.FSharp.Math
open System.Text
open System
open System.Threading
open AMSR

let inline show_usage() = 
        print_endline("Usage : AMSR-translator.exe <input_file_path> [<state_space_limit>] [-u]")
        print_endline("")
        print_endline("-u       =   export unique prism model")

let inline translate_to_prism_language(filename: string, limit:int) =
    let filename_root = filename.Substring(0, filename.LastIndexOf("."))
    let myModel = AMSR.AMSR_Parser.AMSR_Parser.Parse(filename)
    myModel.state_space_limit <- limit
    print_endline("Model in " + filename + " parsed.")
    let myLTS = myModel.toLTS()
    print_endline("LTS computed.")
    let myIMC = myModel.toIMC(myLTS)
    print_endline("IMC computed.")
    let myDelimitedIMC = Model.cutIMC(myIMC)
    print_endline("Delimited IMC computed.")
    
    //let myMDPProperties = Model.toMDPProperties(myDelimitedIMC , myModel.properties)
    let myMDPProperties = Model.toMDPProperties_directly(myDelimitedIMC , myModel.properties)
    let myMDP = new mdp(myDelimitedIMC)
        
    //save model file
    let output_filename = filename_root + ".pm"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    
    let prism_model = myMDP.ToPrismModel(myModel.symbols)
    
    tex.Write(prism_model)//myMDP)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".pm generated." )
    
    //save properties file
    let output_filename = filename_root + ".pctl"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(myMDPProperties)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".pctl generated." )
    (*
    //save graphical representation of LTS
    Model.saveLTSGraphImage(myLTS,filename_root + ".LTS.graph.png")
    print_endline(filename_root + ".LTS.graph.png generated.")
    
    //save graphical representation of IMC
    Model.saveGraphImage(myDelimitedIMC,filename_root + ".IMC.graph.png")
    print_endline(filename_root + ".IMC.graph.png generated.")
    
    //save graphical representation of MDP
    Model.saveMDPGraphImage(myMDP,filename_root + ".MDP.graph.png")
    print_endline(filename_root + ".MDP.graph.png generated.")
    *)
    ()

let inline translate_to_prism_transitions_and_states(filename: string, limit:int) =
    let filename_root = filename.Substring(0, filename.LastIndexOf("."))
    let myModel = AMSR.AMSR_Parser.AMSR_Parser.Parse(filename)
    myModel.state_space_limit <- limit
    print_endline("Model in " + filename + " parsed.")
    let myLTS = myModel.toLTS()
    print_endline("LTS computed.")
    let myIMC = myModel.toIMC(myLTS)
    print_endline("IMC computed.")
    let myDelimitedIMC = Model.cutIMC(myIMC)
    print_endline("Delimited IMC computed.")
    
    let myMDPProperties = Model.toMDPProperties_directly(myDelimitedIMC , myModel.properties)
    let myMDP = new mdp(myDelimitedIMC)
    
    let a,b,c = myMDP.ToPRISMTransitionsAndStates(myModel.symbols)
    
    //save state file
    let output_filename = filename_root + ".sta"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(a)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".sta generated." )
    
    
    //save transitions file
    let output_filename = filename_root + ".tra"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(b)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".tra generated." )
    
     
    //save initial state file
    let output_filename = filename_root + ".ini"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(c)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".ini generated." )
    
    
    //save properties file
    let output_filename = filename_root + ".pctl"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(myMDPProperties)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".pctl generated." )
    
    
    //save graphical representation of LTS
    Model.saveLTSGraphImage(myLTS,filename_root + ".LTS.graph.png")
    print_endline(filename_root + ".LTS.graph.png generated.")
    
    //save graphical representation of IMC
    Model.saveGraphImage(myDelimitedIMC,filename_root + ".IMC.graph.png")
    print_endline(filename_root + ".IMC.graph.png generated.")
    
    //save graphical representation of MDP
    Model.saveMDPGraphImage(myMDP,filename_root + ".MDP.graph.png")
    print_endline(filename_root + ".MDP.graph.png generated.")
    
    ()

let inline go() =
    let args = System.Environment.GetCommandLineArgs()
    let export_unique_file = 
        Seq.exists (fun el -> el = "-u") args
    if export_unique_file 
        then
            match args.Length
                with 
                    | 3 -> translate_to_prism_language(args.[1], 10000)
                    | 4 ->  translate_to_prism_language(args.[1], System.Int32.Parse(args.[2]))
                    | n -> 
                        show_usage()
        else    
            match args.Length
                with 
                    | 2 ->  translate_to_prism_transitions_and_states(args.[1], 10000)
                        
                    | 3 ->  translate_to_prism_transitions_and_states(args.[1], System.Int32.Parse(args.[2]))
                    | n ->  show_usage()
    ()
    (*
let test_multisets() =
    let a = new multiset(4);
    let b = new multiset(4);
    b.add2(0,5); b.add2(1,5)
    a.add2(0,5); a.add2(1,5)
    let mutable myset = Set.Empty
    myset <- myset.Add(a)
    let result = myset.Contains(b)
    Console.WriteLine(result) ///OK
    ()
*)
let _ =
    /// Measures the time it takes to execute a function
    let timer = System.Diagnostics.Stopwatch.StartNew ()
    //test_multisets()
    go()
    (*try 
        go()
    with 
            | e -> ()   
    *)
    printfn "Time taken = %O" timer.Elapsed