using System.Collections;
using System.Collections.Generic;
using System.Text;

using AMSR;

using System;



public class Parser {
	const int _EOF = 0;
	const int _Id = 1;
	const int _Number = 2;
	const int _RULESDIC = 3;
	const int _TERMDIC = 4;
	const int _SPECIESDIC = 5;
	const int _PROPDIC = 6;
	const int maxT = 19;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;
	
	public Scanner scanner;
	public Errors  errors;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;

public static Model model;
public static SymbolTable<string> symbols;
public static List<ab_rw_rule> rulesList;
public static List<property> PropertiesList;



	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (string msg) {
		if (errDist >= minErrDist) errors.SemErr(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}


	bool WeakSeparator(int n, int syFol, int repFol) {
		int kind = la.kind;
		if (kind == n) {Get(); return true;}
		else if (StartOf(repFol)) {return false;}
		else {
			SynErr(n);
			while (!(set[syFol, kind] || set[repFol, kind] || set[0, kind])) {
				Get();
				kind = la.kind;
			}
			return StartOf(syFol);
		}
	}

	
	void AMSR() {
		rulesList = new List<ab_rw_rule>();
		symbols = new SymbolTable<string>();
		PropertiesList = new List<property>();
		StringBuilder TEMP = new StringBuilder(); 
		
		Expect(5);
		species();
		Expect(3);
		rules();
		multiset initialTerm = new multiset(symbols.Count()); 
		Expect(4);
		terms(initialTerm);
		if (la.kind == 6) {
			Get();
			properties();
		}
		model = new Model(rulesList.ToArray(), initialTerm, symbols, PropertiesList.ToArray());	
		
	}

	void species() {
		elementsdic();
	}

	void rules() {
		rule();
		if (la.kind == 1 || la.kind == 2) {
			rules();
		}
	}

	void terms(multiset ms) {
		StringBuilder TEMP = new StringBuilder();  
		term(TEMP);
		
		if (la.kind == 7) {
			Get();
			Expect(2);
		}
		long temp;
		long.TryParse(t.val, out temp);
		if(temp == 0) temp++;
		
		          if (Microsoft.FSharp.Core.Option<int>.get_IsNone(symbols.inverse_get(TEMP.ToString().ToLower()))) {Console.WriteLine("Some species are missing from species declaration."); throw new Exception("Error in parsing"); }
					    ms.add(symbols.inverse_get(TEMP.ToString().ToLower()).Value, (int) temp);
					  
		if (la.kind == 1 || la.kind == 2) {
			terms(ms);
		}
	}

	void properties() {
		property();
		if (StartOf(1)) {
			properties();
		}
	}

	void property() {
		StringBuilder TEMP = new StringBuilder(); 
		StringBuilder TEMP_DECL = new StringBuilder(); 
		
		if (StartOf(2)) {
			Get();
			TEMP_DECL.Append(" " + t.val + " "); 
		} else if (StartOf(3)) {
			relation(TEMP_DECL);
		} else SynErr(20);
		while (StartOf(4)) {
			if (StartOf(5)) {
				Get();
				TEMP_DECL.Append(" " + t.val + " "); 
			} else {
				relation(TEMP_DECL);
			}
		}
		Expect(7);
		property prop = new property();
		prop.decl_string = TEMP_DECL.ToString();
		PropertiesList.Add(prop);
		
	}

	void relation(StringBuilder tt) {
		switch (la.kind) {
		case 8: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		case 9: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		case 10: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		case 11: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		case 12: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		case 13: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		case 14: {
			Get();
			tt.Append(" " + t.val + " ");
			break;
		}
		default: SynErr(21); break;
		}
	}

	void elementsdic() {
		StringBuilder TEMP = new StringBuilder();  
		Identifier(TEMP);
		symbols.add(TEMP.ToString().ToLower());
		if (la.kind == 1 || la.kind == 2) {
			elementsdic();
		}
	}

	void rule() {
		StringBuilder TEMP = new StringBuilder();							
		Identifier(TEMP);
		string name = TEMP.ToString();	TEMP = new StringBuilder();
		Expect(15);
		Expect(16);
		multiset lhs = new multiset(symbols.Count());	
		multiset rhs = new multiset(symbols.Count());
		
		terms(lhs);
		Expect(17);
		if (la.kind == 1 || la.kind == 2) {
			terms(rhs);
		}
		Expect(17);
		rate(TEMP);
		double lwB; 
		double.TryParse(TEMP.ToString(), out lwB); 
		TEMP = new StringBuilder();
		
		Expect(17);
		rate(TEMP);
		double upB; 
		double.TryParse(TEMP.ToString(), out upB); 
		TEMP = new StringBuilder();
		
		Expect(18);
		interval rateInterval = new interval(lwB,upB);
		rulesList.Add(new ab_rw_rule(lhs,rateInterval,rhs));
		
	}

	void Identifier(StringBuilder TEMP) {
		if (la.kind == 2) {
			Get();
			TEMP.Append(t.val); 
		}
		Expect(1);
		TEMP.Append(t.val.ToLower()); 
	}

	void rate(StringBuilder TEMP) {
		Expect(2);
		TEMP.Append(t.val); 
	}

	void term(StringBuilder TEMP) {
		Identifier(TEMP);
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		AMSR();

    Expect(0);
	}
	
	bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x},
		{x,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, x},
		{x,T,T,T, T,T,T,T, x,x,x,x, x,x,x,T, T,T,T,T, x},
		{x,x,x,x, x,x,x,x, T,T,T,T, T,T,T,x, x,x,x,x, x},
		{x,T,T,T, T,T,T,x, T,T,T,T, T,T,T,T, T,T,T,T, x},
		{x,T,T,T, T,T,T,x, x,x,x,x, x,x,x,T, T,T,T,T, x}

	};
} // end Parser


public class Errors {
	public int count = 0;                                    // number of errors detected
	public System.IO.TextWriter errorStream = Console.Out;   // error messages go to this stream
  public string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
  
	public void SynErr (int line, int col, int n) {
		string s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "Id expected"; break;
			case 2: s = "Number expected"; break;
			case 3: s = "RULESDIC expected"; break;
			case 4: s = "TERMDIC expected"; break;
			case 5: s = "SPECIESDIC expected"; break;
			case 6: s = "PROPDIC expected"; break;
			case 7: s = "\"*\" expected"; break;
			case 8: s = "\"<=\" expected"; break;
			case 9: s = "\"=<\" expected"; break;
			case 10: s = "\"<\" expected"; break;
			case 11: s = "\"=>\" expected"; break;
			case 12: s = "\">=\" expected"; break;
			case 13: s = "\"=\" expected"; break;
			case 14: s = "\">\" expected"; break;
			case 15: s = "\":\" expected"; break;
			case 16: s = "\"(\" expected"; break;
			case 17: s = "\",\" expected"; break;
			case 18: s = "\")\" expected"; break;
			case 19: s = "??? expected"; break;
			case 20: s = "invalid property"; break;
			case 21: s = "invalid relation"; break;

			default: s = "error " + n; break;
		}
		errorStream.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void SemErr (int line, int col, string s) {
		errorStream.WriteLine(errMsgFormat, line, col, s);
		count++;
	}
	
	public void SemErr (string s) {
		errorStream.WriteLine(s);
		count++;
	}
	
	public void Warning (int line, int col, string s) {
		errorStream.WriteLine(errMsgFormat, line, col, s);
	}
	
	public void Warning(string s) {
		errorStream.WriteLine(s);
	}
} // Errors


public class FatalError: Exception {
	public FatalError(string m): base(m) {}
}

