﻿#light

open System
open System.Collections
open System.Collections.Generic
open System.IO
open Microsoft.FSharp.Math
open System.Text
open System
open System.Threading
open AMSR
open myMath
open ZedGraph

let inline readFileToString(filename:string) =
    let  rdr = File.OpenText(filename)
    rdr.ReadToEnd()

let inline execute(command,par,filename:string) =
    // Start the child process.
    try 
        
        //Create process
        let pProcess = new System.Diagnostics.Process();

        //strCommand is path and file name of command to run
        pProcess.StartInfo.FileName <- command;

        //strCommandParameters are parameters to pass to program
        pProcess.StartInfo.Arguments <- par;

        pProcess.StartInfo.UseShellExecute <- false;

        //Set output of program to be written to process output stream
        pProcess.StartInfo.RedirectStandardOutput <- true;   

        //Optional
        pProcess.StartInfo.WorkingDirectory <- Environment.CurrentDirectory //+ "\prism-3.1.1\prism\bin";

        print_endline("Process :"+ pProcess.StartInfo.FileName)

        //Start the process
        pProcess.Start() |> ignore

        //Get program output
        let strOutput = pProcess.StandardOutput.ReadToEnd();

        //Wait for process to finish
        pProcess.WaitForExit();
        
        //get results
        let results_filename = (*( filename.Substring(0, filename.LastIndexOf("."))  )*)filename+"_results.txt"

        AMSR.Prism_Parser.Prism_Parser.Parse(results_filename)
    with
        | e -> Console.WriteLine("ERROR IN PRISM INVOCATION"); new List<float>();
        
        

let inline executePrismMono(model_name:string) =
    //run prism
    let model_name = ( model_name.Substring(0, model_name.LastIndexOf("."))  )
    let strCommand = "sh prism"
    let strCommandParameters = model_name+".pm "+ model_name+ ".pctl -exportresults "+model_name+"_results.txt"
    execute(strCommand,strCommandParameters,model_name)


//let wrapPrism(inputmodelpath, propertiespath, outputpath, properties, symbols) = 
let inline executePrismDotNET(model_name:string) = 
    //run prism
    let model_name = ( model_name.Substring(0, model_name.LastIndexOf("."))  )
    let strCommand = "prism.bat"
    let strCommandParameters = model_name+".pm "+ model_name+ ".pctl -exportresults "+model_name+"_results.txt"
    execute(strCommand,strCommandParameters,model_name)


let inline executePrismMono_on_model(model_name:string) =
    //run prism
    let model_name = ( model_name.Substring(0, model_name.LastIndexOf("."))  )
    let strCommand = "sh prism"
    let strCommandParameters = model_name+".pm "+ model_name+ ".pctl -exportresults "+model_name+"_results.txt"
    execute(strCommand,strCommandParameters,model_name)


//let wrapPrism(inputmodelpath, propertiespath, outputpath, properties, symbols) = 
let inline executePrismDotNET_on_model(model_name:string) = 
    //run prism
    let model_name = ( model_name.Substring(0, model_name.LastIndexOf("."))  )
    let strCommand = "prism.bat"
    let initString = System.IO.File.ReadAllText(model_name + ".ini")
    let maxiters = 100000
    let strCommandParameters = "-importtrans " + model_name+".tra -mdp -fixdl " + "-importstates " + model_name+".sta "+ "-importinit " + initString + " " + model_name+ ".pctl -maxiters " + maxiters.ToString() + " -exportresults "+model_name+"_results.txt"
    
    let a = strCommand + " " + strCommandParameters
    print_endline(a.ToString())
    
    execute(strCommand,strCommandParameters,model_name)


let inline show_usage() = 
        print_endline("Usage : AMSR-translator.exe <input_file_path> ")

let inline translate_via_prism_language(filename: string, limit:int) =
    let filename_root = filename.Substring(0, filename.LastIndexOf("."))
    let myModel = AMSR.AMSR_Parser.AMSR_Parser.Parse(filename)
    myModel.state_space_limit <- limit  
    let myLTS = myModel.toLTS()
    let myIMC = myModel.toIMC(myLTS)
    let myDelimitedIMC = Model.cutIMC(myIMC)
    let myMDPProperties = Model.toMDPProperties(myDelimitedIMC , myModel.properties)
    let myMDP = new mdp(myDelimitedIMC)
    
    
    if Model.DEBUG 
        then
        print_endline("MODEL")
        print_endline(myModel.ToString())
        print_endline("LT")
        print_endline(myLTS.ToString())
        print_endline("IMC")
        print_endline(myIMC.ToString())
        print_endline("DELIMITED IMC")
        print_endline(myDelimitedIMC.ToString())
        
    
    //save model file
    let output_filename = filename_root + ".pm"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(myMDP)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".pm generated." )
    
    //save properties file
    let output_filename = filename_root + ".pctl"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(myMDPProperties)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".pctl generated." )
    
    myModel.properties
    
let inline translate_via_prism_model(filename: string, limit:int) =
    let filename_root = filename.Substring(0, filename.LastIndexOf("."))
    let myModel = AMSR.AMSR_Parser.AMSR_Parser.Parse(filename)
    myModel.state_space_limit <- limit
    print_endline("Model in " + filename + " parsed.")
    let myLTS = myModel.toLTS()
    print_endline("LTS computed.")
    let myIMC = myModel.toIMC(myLTS)
    print_endline("IMC computed.")
    let myDelimitedIMC = Model.cutIMC(myIMC)
    print_endline("Delimited IMC computed.")
    
    if Model.DEBUG 
        then
        print_endline("MODEL")
        print_endline(myModel.ToString())
        print_endline("LT")
        print_endline(myLTS.ToString())
        print_endline("IMC")
        print_endline(myIMC.ToString())
        print_endline("DELIMITED IMC")
        print_endline(myDelimitedIMC.ToString())
    
    let myMDPProperties = Model.toMDPProperties_directly(myDelimitedIMC , myModel.properties)
    let myMDP = new mdp(myDelimitedIMC)
    
    let (a:string),(b:string),(c:string) = myMDP.ToPRISMTransitionsAndStates(myModel.symbols)
    
    //save state file
    let output_filename = filename_root + ".sta"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(a)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".sta generated." )
    
    
    //save transitions file
    let output_filename = filename_root + ".tra"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(b)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".tra generated." )
    
     
    //save initial state file
    let output_filename = filename_root + ".ini"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(c)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".ini generated." )
    
    
    //save properties file
    let output_filename = filename_root + ".pctl"
    let inline open_file(filename:string) =
        let (t:FileInfo) = new FileInfo(filename);
        let  (tex:StreamWriter) =t.CreateText();
        tex
    let tex = open_file(output_filename)
    tex.Write(myMDPProperties)
    tex.Flush();
    tex.Close();
    print_endline(filename_root + ".pctl generated." )
    
    myModel.properties

(*
//buono se TI e TN sono intervallati
let makeAndSavePlot(filename:string, properties: property array, results: _ List) =
    let zg1 = new ZedGraph.ZedGraphControl();
    let myPane = zg1.GraphPane //new GraphPane( new System.Drawing.RectangleF( float32 0.0, float32  0.0, float32 1280.0, float32 1024.0 ),"Link Distribution", "X Title", "Y Title" );
    //myPane.Rect <- new System.Drawing.RectangleF( float32 0.0, float32  0.0, float32 1280.0, float32 1024.0 )
    
    
    //ColorSymbolRotator 
    let min_list = new PointPairList();
    let max_list = new PointPairList();
    
    myPane.Title.Text <- "Links Distribution"
    let mutable j = 0
    if results.Count > 0 
                        then
                            for i in [0.. 2 .. properties.Length-1] do
                               
                                let up, low = results.[j], results.[j+1]
                                //let lowr, upr = new Rational(low), new Rational(up)
                                //out.AppendLine("Reach("+properties.[i].decl_string+") = ["+ if low = 1.0 then low.ToString() else (new Rational(low)).ToString() + " , "+ if up = 1.0 then low.ToString() else (new Rational(up)).ToString()+ "]")|>ignore
                                let interval = new interval(low,up)
                                
                                
                                //out.AppendLine("Reach( "+properties.[i].decl_string+" ) = "+  interval.ToString() + " = " + interval.ToFractionString())|>ignore
                                min_list.Add(float properties.[i].value ,low)
                                max_list.Add(float properties.[i].value ,up)
                                
                                j <- j+4
                            done
                            
    
    let minCurve = myPane.AddCurve("min", min_list, System.Drawing.Color.Blue);
    minCurve.Line.Width <- Float.to_float32 3.0
    let maxCurve = myPane.AddCurve("max", max_list, System.Drawing.Color.Red);
    maxCurve.Line.Width <- Float.to_float32 3.0
    
    myPane.Title.Text <-  filename
    
    myPane.XAxis.Title.Text <- "TI";
    myPane.YAxis.Title.Text <- "p"; 
    
    let bm = new System.Drawing.Bitmap(1, 1);
    let g = System.Drawing.Graphics.FromImage(bm);
    zg1.GraphPane.AxisChange(g);
    zg1.Invalidate();
    //myPane.AxisChange();
    myPane.Legend.IsVisible <- false;
    // For ZedGraph 4.3, the next line is: myPane.Image.Save( @"zedgraph.png", ImageFormat.Png );
    myPane.GetImage(1280, 1024, float32 32.0, true).Save(filename + ".result.png", System.Drawing.Imaging.ImageFormat.Png);
    //myPane.GetImage().Save(filename, System.Drawing.Imaging.ImageFormat.Png);
    ()

*)

///buono per 12 * 6
let makeAndSavePlot(filename:string, properties: property array, results: _ List, limit) =
    
    let nr_blocks =  2
    let nr_properties_for_block = 18
    
    let titles =
        [|
        //"TN >= value";
        "TN = value";
        //"TN <= value";
        //"TI >= value";
        "TI = value";
        //"TI <= value";
        |]
        

    //let js = [|0, 24, |]    
    let mutable j =  0// ((((block_count) - 1) * nr_properties_for_block * 2)  ) 
    for block_count = 1 to nr_blocks do

        //ColorSymbolRotator 
        let min_list = new PointPairList();
        let max_list = new PointPairList();
        
        let zg1 = new ZedGraph.ZedGraphControl();
        let myPane = zg1.GraphPane //new GraphPane( new System.Drawing.RectangleF( float32 0.0, float32  0.0, float32 1280.0, float32 1024.0 ),"Link Distribution", "X Title", "Y Title" );
        //myPane.Rect <- new System.Drawing.RectangleF( float32 0.0, float32  0.0, float32 1280.0, float32 1024.0 )
    
        
        myPane.Title.Text <- "Links Distribution"
        //let mutable j =  ((((block_count) - 1) * nr_properties_for_block * 2)  ) 
        if results.Count > 0 
                            then
                                for i in [((nr_properties_for_block * (block_count - 1)) ).. 1 .. ((nr_properties_for_block * (block_count - 1)) + nr_properties_for_block - 1)] do
                                   
                                    let up, low = results.[j], results.[j+1]
                                    //let lowr, upr = new Rational(low), new Rational(up)
                                    //out.AppendLine("Reach("+properties.[i].decl_string+") = ["+ if low = 1.0 then low.ToString() else (new Rational(low)).ToString() + " , "+ if up = 1.0 then low.ToString() else (new Rational(up)).ToString()+ "]")|>ignore
                                    let interval = new interval(low,up)
                                    
                                    //print_endline(j.ToString() ^ " - " ^ i.ToString())
                                    
                                    //out.AppendLine("Reach( "+properties.[i].decl_string+" ) = "+  interval.ToString() + " = " + interval.ToFractionString())|>ignore
                                    min_list.Add(float properties.[i].value ,low)
                                    max_list.Add(float properties.[i].value ,up)
                                    
                                    j <- j+2
                                done
     
        let minCurve = myPane.AddCurve("min", min_list, System.Drawing.Color.Blue);
        minCurve.Line.Width <- float32 3.0
        let maxCurve = myPane.AddCurve("max", max_list, System.Drawing.Color.Red);
        maxCurve.Line.Width <- float32 3.0
        
        myPane.Title.Text <-  titles.[block_count - 1] + " @ " + limit.ToString() + " -- " + filename
        
        myPane.XAxis.Title.Text <- "value";
        myPane.YAxis.Title.Text <- "p"; 
        
        let bm = new System.Drawing.Bitmap(1, 1);
        let g = System.Drawing.Graphics.FromImage(bm);
        zg1.GraphPane.AxisChange(g);
        zg1.Invalidate();
        //myPane.AxisChange();
        myPane.Legend.IsVisible <- false;
        // For ZedGraph 4.3, the next line is: myPane.Image.Save( @"zedgraph.png", ImageFormat.Png );
        myPane.GetImage(1280, 1024, float32 32.0, true).Save(filename + "." + block_count.ToString() + "@" + limit.ToString() + ".result.png", System.Drawing.Imaging.ImageFormat.Png);                           
    done
    
    //myPane.GetImage().Save(filename, System.Drawing.Imaging.ImageFormat.Png);
    ()
let inline go() =
    let args = System.Environment.GetCommandLineArgs()
    match args.Length
            with 
                | 2 | 3 ->
                    let limit = if args.Length = 2 then 10000 else System.Int32.Parse(args.[2])  
                    let filename = args.[1]
                    let properties = translate_via_prism_model(filename, limit)
                    let filename_root = filename.Substring(0, filename.LastIndexOf("."))
                    //detect runtime
                    let mutable mono = false
                    let t = System.Type.GetType ("Mono.Runtime")
                    let tt = System.Type.GetType ("Runtime")
                    in match t with
                        | null ->()
                        | _ -> mono <- true
    
                    //lauch prism
                    let prismtimer = System.Diagnostics.Stopwatch.StartNew ()
                    let results = if mono then executePrismMono_on_model(filename) else executePrismDotNET_on_model(filename)  
                    printfn "PRISM Time taken = %O" prismtimer.Elapsed
                    let out = new StringBuilder("")
                    let mutable j = 0
                    if results.Count > 0 
                        then
                            for i in [0..properties.Length-1] do
                                let up, low = results.[j], results.[j+1]
                                //let lowr, upr = new Rational(low), new Rational(up)
                                //out.AppendLine("Reach("+properties.[i].decl_string+") = ["+ if low = 1.0 then low.ToString() else (new Rational(low)).ToString() + " , "+ if up = 1.0 then low.ToString() else (new Rational(up)).ToString()+ "]")|>ignore
                                let interval = new interval(low,up)
                                out.AppendLine("Reach( "+properties.[i].decl_string+" ) = "+  interval.ToString() + " = " + interval.ToFractionString())|>ignore
                                
                                j <- j+2
                            done
                    print_endline(out.ToString())
                    
                      //save model file
                    let output_filename = filename_root + ".results.txt"
                    let open_file(filename:string) =
                        let (t:FileInfo) = new FileInfo(filename);
                        let  (tex:StreamWriter) =t.CreateText();
                        tex
                    let tex = open_file(output_filename)
                    tex.Write(out.ToString())
                    tex.Flush();
                    tex.Close();
                    print_endline(filename_root + ".results.txt generated." )
                    
                      //create the plot of probabilities
                    //if filename.Contains("Tumor") then makeAndSavePlot(filename,properties,results, limit) 
                    
                | n -> show_usage(); failwith "";
    ()

let _ =
    /// Measures the time it takes to execute a function
    let timer = System.Diagnostics.Stopwatch.StartNew ()
    
    try 
        go()
    with 
        | e -> (print_endline("Error : " + e.Message))   
    
    printfn "Time taken = %O" timer.Elapsed