﻿#light

namespace AMSR

open System.Collections.Generic;
open System.Text;
open Microsoft.FSharp.Math;
open myMath;

module MyMath =
    begin
    /// The constant pi = 3.141596...
    let PI = System.Math.PI
/// The constant e = 2.718281828
    let E = System.Math.E
    let Gamma_n = 10
    let Gamma_r = 10.900511
    let Gamma_dk = [| 2.48574089138753565546e-5;
                      1.05142378581721974210;
                      -3.45687097222016235469;
                      4.51227709466894823700;
                      -2.98285225323576655721;
                      1.05639711577126713077;
                      -1.95428773191645869583e-1;
                      1.70970543404441224307e-2;
                      -5.71926117404305781283e-4;
                      4.63399473359905636708e-6;
                      -2.71994908488607703910e-9 |]

    /// Computes the logarithm of the gamma function for real numbers.
    let GammaLn z =    
        /// This method computes the logarithm of the gamma function in the upper half plane.
        let gammaln_p z =
            let mutable s = Gamma_dk.[0]
            for i=1 to Gamma_n do
                s <- s + Gamma_dk.[i] / (z + (float (i-1)))
            (log s) + (log 2.0) + 0.5 * (log (E/PI)) + (z - 0.5) * (log ((z - 0.5 + Gamma_r)/E))

        if z < 0.5 then                                     // TODO how do we choose when to mirror?
            log PI - log (sin (PI*z)) - gammaln_p(1.0-z)
        else
            gammaln_p z

/// Computes the natural logarithm of the factorial function.
    let FactorialLn (x: int) : float =
        if x < 0 then failwith "Log factorial not defined for n < 0"
        if x <= 1 then 0.0 else GammaLn ((float x) + 1.0)

/// Computes the binomial coefficient.
    let Binomial (n: int) (h: int) = floor (0.5 + exp ((FactorialLn n) - (FactorialLn h) - (FactorialLn (n-h))))

end
(*
open Microsoft.Glee
open Microsoft.Glee.GraphViewerGdi
open Microsoft.Glee.Drawing
*)
/// a generic SymbolTable
/// that map element of type 'a (identifier) in element of type int (addresses)
type 'a SymbolTable =  
    class
     // FIELDS
        val mutable table :  Dictionary< int, 'a > 
        val mutable inverse_table : Dictionary< 'a, int >
    // CONSTRUCTORS
        new() = { table = new Dictionary<int, 'a>(); inverse_table =  new Dictionary<'a, int>(); }
    // METHODS
        /// get the number of different values
        member inline st.Count() : int = st.table.Count
        /// get a value by an address
        member st.get(i: int) : 'a option =
            let ok, res = st.table.TryGetValue(i)
            if ok then Some(res)
                  else None
        /// get an address by an identifier
        member st.inverse_get(i: 'a) : int option =
            let ok, res = st.inverse_table.TryGetValue(i)
            if ok then Some(res - 1)
                  else None
        /// add a value and get it's address
        member st.add(elem : 'a) : int =
            let ok,res = st.inverse_table.TryGetValue(elem)
            if ok 
                then (  res - 1 )
                else (
                    st.table.Add(st.Count(), elem )
                    st.inverse_table.Add(elem, st.Count())
                    st.Count()-1
                )
        /// remove a value
        member st.remove(elem : 'a) : unit =
            let ok,res = st.inverse_table.TryGetValue(elem)
            if (ok) 
                then (
                    st.inverse_table.Remove(elem) |> ignore
                    st.table.Remove(res) |> ignore
                )
                else ()
        /// get string representation of the symbol table
        override st.ToString() : string =
            let mutable output = new StringBuilder("[ ")
            for keyVal in st.table do 
                output.Append("( " + keyVal.Key.ToString() + " , " + box(keyVal.Value).ToString() + " )") |> ignore 
            output.Append(" ]") |> ignore
            output.ToString()

           
    end


and interval =
    class
        val low : double
        val up : double
        new(l,u) = if l<=u then {low=l; up=u;} else {low=u; up=l;}
        override x.ToString() =
            "[" ^ x.low.ToString() ^ "," ^ x.up.ToString() ^"]"
        member x.ToFractionString() =
            "[" ^ ( if x.low = 0.0 || x.low = 1.0 || x.low = -1.0 then x.low.ToString() else Model.fractionString(x.low) )^ "," 
             ^ ( if x.up = 0.0 || x.up = 1.0 || x.up = -1.0 then x.up.ToString() else  Model.fractionString(x.up))^"]"
        member inline x.get_Item(i:int) = //questo sarebbe da fare in override ??!!
            match i with 
                | 0 -> x.low
                | 1 -> x.up
                | _ -> failwith ""
        static member inline add (i1:interval) (i2:interval) =
            new interval(i1.low + i2.low, i1.up + i2.up)
        override x.GetHashCode() = box(x).GetHashCode()
        override x.Equals(y:obj) =
            match y with 
                |   :? interval as el    -> x.low = el.low && x.up = el.up 
                |   _                   -> false 
        interface System.IComparable with
            member x.CompareTo(y:obj) = 
                match y with
                    |   :? interval as el    -> if x.low = el.low && x.up = el.up  then 0 else 1           
                    |   _                    -> -1
        end
    end
    
 and state =
    class
    val id : int
    new(id) = {id=id;}
    end
 
 and rel =
        | great = 0 
        | geq = 1
        | equal = 2
        | leq = 3
        | less = 4
 
 and property =
    class
    val mutable id : int
    val mutable relation : rel 
    val mutable value : int
    val mutable decl_string : string
    new(id, relation, value, decl_string) = {id=id; relation=relation; value=value; decl_string=decl_string;}
    new() = {id=0; relation=rel.equal; value=0; decl_string=""}
    member x.toPrismProperty() = 
        //find states that satisfies the property
        //generate the prism propoerties
        // Pmax=? [F (state=1) | (state=2]
        ()
    end
   
 and multiset =
    class
    val ar :  int array
    [<OverloadID("ctor1")>]
    new(size) = { ar = Array.create size 0 }
    [<OverloadID("ctor2")>]
    new(array) = {ar=array}
    member x.add(i:int) =
            x.ar.[i] <- x.ar.[i] + 1 
    member x.add(i:int, t:int) =
            x.ar.[i] <- x.ar.[i] + t   
    member inline x.size = x.ar.Length
    member inline x.cardinality = 
            Array.fold (fun acc el -> acc + el) 0 x.ar
    override x.ToString() =
            let out = new StringBuilder("[");
            for i=0 to x.ar.Length - 1 do
                out.Append(" " + x.ar.[i].ToString() ) |> ignore
                //if x.ar.[i] <> 0 then out.Append(" " + i.ToString() + "*"+ x.ar.[i].ToString() ) |> ignore
            done
            out.Append("]").ToString();
            
    member inline x.ToPRISMInitString(symbols: 'a SymbolTable) =
            let out = new StringBuilder();
            for i=0 to x.ar.Length - 1 do

                out.Append(symbols.get(i).Value.ToString() + "="+ x.ar.[i].ToString() + ",") |> ignore
            done
            (if x.ar.Length > 0 then out.Remove(out.Length - 1, 1 )|> ignore)  
            out.Append("").ToString();
    
    member inline x.ToPRISMString() =
            let out = new StringBuilder();
            for i=0 to x.ar.Length - 1 do

                out.Append(x.ar.[i].ToString() + ",") |> ignore
            done
            (if x.ar.Length > 0 then out.Remove(out.Length - 1, 1 )|> ignore)  
            out.Append("").ToString();
            
    member inline x.ToPRISMString2(symbols: string array) =
            let out = new StringBuilder();
            for i=0 to x.ar.Length - 1 do

                out.Append(" (" + symbols.[i] + "=" + x.ar.[i].ToString() + ") &") |> ignore
            done
            (if x.ar.Length > 0 then out.Remove(out.Length - 1, 1 )|> ignore)  
            out.Append("").ToString();
    
    member  x.ToString(symbols:string SymbolTable) =
            let out = new StringBuilder("[");
            for i=0 to x.ar.Length - 1 do

                if x.ar.[i] <> 0 then out.Append(" " + symbols.get(i).Value.ToString() + "*"+ x.ar.[i].ToString()) |> ignore
            done
            out.Append("]").ToString();
            
    static member inline submultiset (m1:multiset) (m2:multiset) = //checks if m1 is a submultiset of m2 
        m1.size = m2.size
        &&
        Array.fold2 (fun acc el1 el2 -> acc && (el1 <= el2)) true m1.ar m2.ar
        
    static member inline add2((m1:multiset),(m2:multiset)) = //return a new multiset equal to m2+m1
        match (m1.size = m2.size) 
            with
                | false -> failwith("");
                | _-> 
                        let out = new multiset(m1.size)
                        for i in [0 .. (m1.size - 1)] 
                            do
                            out.ar.[i] <- m1.ar.[i] + m2.ar.[i]
                            done
                        out   
    static member inline add (m1:multiset) (m2:multiset) = //return a new multiset equal to m2+m1
        match (m1.size = m2.size) 
            with
                | false -> failwith("");
                | _-> 
                        let out = new multiset(m1.size)
                        for i in [0 .. (m1.size - 1)] 
                            do
                            out.ar.[i] <- m1.ar.[i] + m2.ar.[i]
                            done
                        out
                        
    static member inline sub (m1:multiset) (m2:multiset) = //return m2-m1
        match (m1.size = m2.size) 
            with
                | false -> failwith("");
                | _-> 
                        let out = new multiset(m1.size)
                        for i in [0 .. (m1.size - 1)] 
                            do
                            out.ar.[i] <- max (m2.ar.[i] - m1.ar.[i]) 0
                            done
                        out
     override x.Equals(y:obj) =
            match y with 
                |   :? multiset as el    ->(* if*) x.size = el.size && (Array.fold2 (fun acc el1 el2 -> acc && (el1 = el2)) true x.ar el.ar) //then print_endline(x.ToString() + "=" + el.ToString()); true else print_endline(x.ToString() + "!=" + el.ToString()); false 
                |   _                   -> false 
     interface System.IComparable with
            member x.CompareTo(y:obj) = 
                //print_endline("uso compare")
                //if x.Equals(y) then 0 else 1
                
                match y with
                    |   :? multiset as el    -> compare x.ar el.ar(*if   
                                                     x.size = el.size && (Array.fold_left2 (fun acc el1 el2 -> acc && (el1 = el2)) true x.ar el.ar) 
                                                     then 0 else 1      *)                
                    |   _                    -> -1
                    
        end
     (*interface IStructuralHash with
            member x.GetStructuralHashCode(_) =  
                LanguagePrimitives.GenericHash(x.ar)
        end
      *) 
    end
 
 and rw_rule =
    class
    val l: multiset
    val r: multiset
    val k: double
    new(l,k,r) = {l=l; r=r; k=k;}
    end
    
 and ab_rw_rule =
    class
    val l: multiset
    val r: multiset
    val ki: interval
    new(l,ki,r) = {l=l; r=r; ki=ki;}
    override x.ToString() =
        "( " + x.l.ToString() + " ) , " + x.ki.ToString() + " , " + "( " + x.r.ToString() + " )"
    member inline x.ToString(symbols) =
        "( " + x.l.ToString(symbols) + " ) , " + x.ki.ToString() + " , " +  "( " + x.r.ToString(symbols) + " )"
    end
 
 and transition =
    class
    val s : state
    val e : state
    val rate : double
    new(s,e,rate) = {s=s; e=e; rate=rate;}
    end
    
 and ab_LTS_transition =
    class
    val s1 : state
    val s2 : state
    val n : int
    val b : interval
    new(s1,n,b,s2) = {s1=s1; n=n; b=b; s2=s2;}
    end
 
 and ab_LTS_transition_relation =
    class
    val mutable tr : (multiset * int * interval * multiset) Set
    val mutable tr_array : (multiset * int * interval * multiset) array
    new() = {tr= Set.empty; tr_array= [||];}
    member x.add(s1,n,b,s2) = 
        x.tr <- Set.add (s1,n,b,s2) x.tr
    member inline x.getTransitions(s1:multiset, s2:multiset) =
        Set.filter (fun tuple -> let a,b,c,d = tuple in a=s1 && d=s2) x.tr
    member inline x.getTransitionsFrom(s1:multiset)=
        //set di transizioni
        Set.filter (fun tuple -> let a,b,c,d = tuple in a=s1) x.tr
    member inline x.getNext(s1:multiset)=
        //set di stati
        Set.map (fun tuple -> let a,b,c,d = tuple in d) (x.getTransitionsFrom(s1))
    member inline x.getRate(s1:multiset, s2:multiset) =
        Set.fold (fun acc tuple -> let a,b,c,d = tuple in interval.add acc c) (new interval(0.0,0.0)) (x.getTransitions(s1,s2))
    member inline x.getExitRate(s1:multiset) =
        Set.fold (fun acc tuple -> let a,b,c,d = tuple in interval.add acc c) (new interval(0.0,0.0)) (x.getTransitionsFrom(s1))
    override x.ToString() =
        let out = new StringBuilder()
        for t in x.tr 
            do
            let a,b,c,d =  t
            out.Append(a.ToString() + " -- ( " + b.ToString() + " , " + c.ToString() + " ) -->" + d.ToString() + " | \n" ) |> ignore
            done
        out.ToString()
    member inline x.ToIndexedStatesString(S:multiset Set) =
        let indexedStateSpace = Set.to_array S //|> Array.rev
        let out = new StringBuilder()
        for t in x.tr 
            do
            let a,b,c,d =  t
            let index_a = Array.findIndex (fun el -> el = a) indexedStateSpace
            let index_d = Array.findIndex (fun el -> el = d) indexedStateSpace
            out.Append("s" + index_a.ToString() + " -- (" + b.ToString() + " , " + c.ToString() + " ) --> " + "s" + index_d.ToString()  + "\n" ) |> ignore
            done
        out.ToString()
    end
 
 and ab_LTS =
    class
    val mutable S : multiset Set
    val s0 : multiset
    val rules : ab_rw_rule array
    val tr : ab_LTS_transition_relation
    val maximum_values_state : multiset
    new(S,s0,rules,tr,max) = {S=S; s0=s0; rules=rules; tr=tr; maximum_values_state=max;}
    member inline x.getTransitions(s1,s2)= x.tr.getTransitions(s1,s2)
    member inline x.getTransitionsFrom(s1:multiset)= x.tr.getTransitionsFrom(s1)
    member inline x.getNext(s1:multiset)= x.tr.getNext(s1)
    member inline x.getRate(s1:multiset, s2:multiset) = x.tr.getRate(s1,s2)
    member inline x.getExitRate(s1:multiset) = x.tr.getExitRate(s1)
    override x.ToString() =
        //S
        let out = new StringBuilder("S\n")
        let mutable i = 0
        for s in x.S  |> Set.to_array //|> Array.rev
            do 
            out.Append("s"+ i.ToString() + ": " + s.ToString() + "  \n") |> ignore
            i <- i+1
            done
        //s0
        out.Append("\ns0\n" + x.s0.ToString()) |> ignore
        //tr
        out.Append("\ntr\n"+ x.tr.ToIndexedStatesString(x.S))|> ignore
        out.ToString()
    member inline x.ToGLEEGraph() =
            let g = new Microsoft.Glee.Drawing.Graph("graph");
            g.Cluster <- true;
            g.Directed <- true;
            g.GraphAttr.NodeAttr.Padding <- 50.0;
            g.GraphAttr.LayerDirection<- Microsoft.Glee.Drawing.LayerDirection.LR;
            g.GraphAttr.NodeAttr.LineWidth <- 2
            g.GraphAttr.EdgeAttr.LineWidth <- 2
            //g.GraphAttr.NodeAttr.FontName <- "Lucida-Sans"
            
            //g.GraphAttr.NodeAttr.Padding <- 3.0;
            let indexedStates = x.S |> Set.to_array //|> Array.rev
            for i=0 to indexedStates.Length - 1 do
                if indexedStates.[i] = x.s0 then
                    let vv = g.AddNode("s"+i.ToString()^" : "^indexedStates.[i].ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.Red
                    vv.NodeAttribute.LineWidth <- 6
                if indexedStates.[i] = x.maximum_values_state then
                    let vv = g.AddNode("s"+i.ToString()^" : "^indexedStates.[i].ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.LightGreen
                    vv.NodeAttribute.LineWidth <- 6
                let current_trans = x.tr.getTransitionsFrom(indexedStates.[i])
                for tr in current_trans do
                    let a,b,c,d = tr
                    let i_d = Array.findIndex (fun el -> el = d) indexedStates
                    let v = g.AddEdge("s"+i.ToString()^" : "^indexedStates.[i].ToString(), "R"^(b+1).ToString() ^ " : " ^ c.ToFractionString() ,"s"+i_d.ToString()^" : "^indexedStates.[i_d].ToString()) //|> ignore // directed inverted
                    //v.EdgeAttr.Color <- Color.Gray
                    ()
                done
                ()
            done
            (*
            for vertex=0 to x.net_in.Length - 1 do
                if x.net_presence.[vertex] then
                    let links = x.net_in.[vertex]
                    for l in links do
                        //let v = g.AddEdge(vertex.ToString(), l.ToString()) //|> ignore //directed
                        let v = g.AddEdge(l.ToString(),vertex.ToString()) //|> ignore // directed inverted
                        v.EdgeAttr.Color <- Color.Gray
                    done
            done
            *)
            g
    end
    
    
 and markov_chain =
    class
    end
 
 and ab_MC_transition_relation =
    class
    val mutable tr : (multiset * interval * multiset) Set
    new() = {tr= Set.empty;}
    member inline x.add(s1,ip,s2) = 
        x.tr <- Set.add (s1,ip,s2) x.tr
    member inline x.getTransitionsFrom(s1:multiset)=
        //set di transizioni
        Set.filter (fun tuple -> let a,b,c = tuple in a=s1) x.tr
    member inline x.getNext(s1:multiset)=
        //set di stati
        Set.map (fun tuple -> let a,b,c = tuple in c) (x.getTransitionsFrom(s1))
    override x.ToString() =
        let out = new StringBuilder()
        for t in x.tr 
            do
            let a,b,c =  t
            out.Append(a.ToString() + " | " + b.ToFractionString() + " | " + c.ToString() + " | \n" ) |> ignore
            //out.Append(a.ToString() + " | " + b.ToString() + " | " + c.ToString() + " | \n" ) |> ignore    
            done
        out.ToString()
    member inline x.ToIndexedStatesString(S:multiset Set) =
        let indexedStateSpace = Set.to_array S //|> Array.rev
        let out = new StringBuilder()
        for t in x.tr 
            do
            let a,b,c =  t
            let index_a = Array.findIndex (fun el -> el = a) indexedStateSpace
            let index_d = Array.findIndex (fun el -> el = c) indexedStateSpace
            out.Append("s" + index_a.ToString() + " -- " + b.ToString()(*b.ToFractionString()*) + " --> " + "s" + index_d.ToString()  + " \n" ) |> ignore
            //out.Append(a.ToString() + " | " + b.ToFractionString() + " | " + c.ToString() + " | \n" ) |> ignore
            //out.Append(a.ToString() + " | " + b.ToString() + " | " + c.ToString() + " | \n" ) |> ignore    
            done
        out.ToString()
    end
 
 
 and ab_markov_chain =
    class
    val mutable S : multiset Set
    val s0 : multiset
    val rules : ab_rw_rule array
    val private ptr : ab_MC_transition_relation
    val mutable indexedPtr : (int * interval * int) Set
    val indexedStateSpace : multiset array
    val mutable maximum_values_state : int
    new(S,s0,rules,ptr, max:multiset) as x 
                        = {
                            S=S;
                            s0=s0; 
                            rules=rules; 
                            ptr=ptr; 
                            indexedStateSpace = (*Array.rev*) (Set.to_array S);
                            indexedPtr = Set.empty;
                            maximum_values_state = -1;
                          }
                          then
                            x.updateIndexedStateSpace()
                            if x.S.Contains max then x.maximum_values_state <- x.index max
                                                else x.maximum_values_state <- x.index x.S.MaximumElement
                                                       
    new(input:ab_markov_chain, myPtr) as x = 
                        {
                            S=input.S;
                            s0=input.s0; 
                            rules=input.rules; 
                            ptr=myPtr; 
                            indexedStateSpace = (*Array.rev*) (Set.to_array input.S);
                            indexedPtr = Set.empty;
                            maximum_values_state = input.maximum_values_state;
                          } //new ab_markov_chain(input, myPtr )
                            then
                            x.updateIndexedStateSpace()
                          
     
    //member x.addToPtr() = ()
    
    member x.getPtr() = x.ptr 
    member private x.updateIndexedStateSpace() =
        x.indexedPtr <-
                                        let maps_states_to_indexes  
                                            =             
                                            fun el ->
                                                let s1,b,s2 = el
                                                let index_a = Array.findIndex (fun el -> el = s1) x.indexedStateSpace
                                                let index_d = Array.findIndex (fun el -> el = s2) x.indexedStateSpace
                                                (index_a, b , index_d)
                                        in  (Set.map maps_states_to_indexes x.ptr.tr)                      
    member x.index s = Array.findIndex (fun el -> el = s) x.indexedStateSpace
    member inline x.get_i_transitions_from(s1:int) =
            Set.filter (fun (ss1,b,s2) -> ss1 = s1) x.indexedPtr
            
    member inline x.get_i_transitions(s1:int, s2:int) =
            Set.filter (fun (ss1,b,ss2) -> ss1 = s1 && ss2 = s1) x.indexedPtr
        
    member inline x.get_i_Next(s1:int) =
            Set.map (fun tuple -> let a,b,c = tuple in c) (x.get_i_transitions_from(s1))
            
    override x.ToString() =
        //S
        let out = new StringBuilder("S\n")
        let mutable i = 0
        for s in x.S |> Set.to_array //|> Array.rev
            do 
            out.Append("s"+ i.ToString() + ": " + s.ToString() + "  \n") |> ignore
            i <- i+1
            done
        //s0
        out.Append("\ns0\n" + x.s0.ToString()) |> ignore
        //tr
        out.Append("\ntr\n"+ x.ptr.ToIndexedStatesString(x.S))|> ignore
        out.ToString()
    
      member x.ToGLEEGraph() : Microsoft.Glee.Drawing.Graph=
            let g = new Microsoft.Glee.Drawing.Graph("graph");
            g.Cluster <- true;
            g.Directed <- true;
            g.GraphAttr.NodeAttr.Padding <- 50.0;
            g.GraphAttr.LayerDirection<- Microsoft.Glee.Drawing.LayerDirection.LR;
            g.GraphAttr.NodeAttr.LineWidth <- 2
            g.GraphAttr.EdgeAttr.LineWidth <- 2
            //g.GraphAttr.NodeAttr.FontName <- "Lucida-Sans"
            
            //g.GraphAttr.NodeAttr.Padding <- 3.0;
            let indexedStates = x.S |> Set.to_array //|> Array.rev
            for i=0 to indexedStates.Length - 1 do
                if indexedStates.[i] = x.s0 then
                    let vv = g.AddNode("s"+i.ToString()^" : "^indexedStates.[i].ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.Red
                    vv.NodeAttribute.LineWidth <- 6
                if x.maximum_values_state > -1 && i = x.maximum_values_state then
                    let vv = g.AddNode("s"+i.ToString()^" : "^indexedStates.[i].ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.LightGreen
                    vv.NodeAttribute.LineWidth <- 6
                let current_trans = x.ptr.getTransitionsFrom(indexedStates.[i])
                for tr in current_trans do
                    let a,b,c = tr
                    let i_c = Array.findIndex (fun el -> el = c) indexedStates
                    let v = g.AddEdge("s"+i.ToString()^" : "^indexedStates.[i].ToString(), b.ToFractionString() ,"s"+i_c.ToString()^" : "^c.ToString()) //|> ignore // directed inverted
                    //v.EdgeAttr.Color <- Color.Gray
                    ()
                done
                ()
            done
            g
    end
    
and mdp_alternative =
    class
    val prob : float
    val s1 : int
    new(p, s) = {prob = p; s1 = s;}
    static member inline difference_threashold = 0.00000000001
    override x.GetHashCode() = box(x).GetHashCode()
    override x.Equals(y:obj) =
            match y with 
                |   :? mdp_alternative as el    ->  x.prob - el.prob <= mdp_alternative.difference_threashold && x.s1 = el.s1 
                |   _                   -> false 
        interface System.IComparable with
            member x.CompareTo(y:obj) = 
                match y with
                    |   :? mdp_alternative as el    -> if x.prob - el.prob <= mdp_alternative.difference_threashold && x.s1 = el.s1  then 0 else 1           
                    |   _                    -> -1
        end
    end

    
and mdp_transition =
    class
    val label : string
    val s : int
    val alternatives : mdp_alternative Set
    new(l, s, a) = {label = l; s = s; alternatives = a}
    override x.GetHashCode() = box(x).GetHashCode()
    override x.Equals(y:obj) =
            match y with 
                |   :? mdp_transition as el    -> (*x.label = el.label &&*)  x.s = el.s && x.alternatives = el.alternatives 
                |   _                   -> false 
        interface System.IComparable with
            member x.CompareTo(y:obj) = 
                match y with
                    |   :? mdp_transition as el    -> if (*x.label = el.label &&*)  x.s = el.s && x.alternatives = el.alternatives   then 0 else 1           
                    |   _                    -> -1
        end
    end
    
 

 
 and mdp =
    class
    val mutable transitions : mdp_transition Set
    val mutable S : int Set
    val mutable multisetsS: multiset Set
    val mutable s0 : multiset
    val mutable maximum_values_state : int
    new() = {transitions = Set.empty; S = Set.empty; multisetsS = Set.empty; s0 = new multiset(0); maximum_values_state = -1;}
    new(input:ab_markov_chain) = 
                                let (m:mdp) = mdp.AB_MC2MDP(input) in
                                {transitions = m.transitions; S = m.S; multisetsS = input.S; s0 = input.s0; maximum_values_state=input.maximum_values_state}
    new(transitions, S, msS, s0, max) =  {transitions = transitions; S = S;  multisetsS = msS; s0 = s0; maximum_values_state = max;}
    
    static member private AB_MC2MDP(input:ab_markov_chain) =
        let result = ref Set.empty
        
        let inline cutMins(l:(int * interval * int) Set, acc:double) =
            let out = ref Set.empty
            for t in l do
                let a,b,c = t
                let k =
                    l |>
                    Set.filter (fun el -> not (el = t))  |>  ///rivedere qui
                    //Set.filter (fun el -> let s1,b,s2 = el in not (s1 = s2)) l |>  ///rivedere qui
                    Set.fold (fun acc el -> let s1,b,s2 = el in acc + b.up) acc 
                    
                let b1 = new interval(max (b.low) (1.0 - k), b.up)
                let t1 = (a,b1,c)
                out := (!out).Add(a,b1,c)
            done
            !out
        
        let inline cutMaxs(l:(int * interval * int) Set, acc:double) =
            let out = ref Set.empty
            for t in l do
                let a,b,c = t
                let k =
                    l |>
                    Set.filter (fun el -> not (el = t))  |>  ///rivedere qui
                    //Set.filter (fun el -> let s1,b,s2 = el in not (s1 = s2)) l |>  ///rivedere qui
                    Set.fold (fun acc el -> let s1,b,s2 = el in acc + b.low) acc 
                    
                let b1 = new interval(b.low ,min (b.up) (1.0 - k))
                let t1 = (a,b1,c)
                out := (!out).Add(a,b1,c)
            done
            !out
        
        let inline sumTo1(input:(double * int) Set) = 
            1.0 = Set.fold (fun acc (b,c) -> acc + b ) 0.0 input
             
        
        let rec pippo(l:(int * interval * int) Set, acc:double, partial_res:mdp_alternative Set, sel:int) =
            if l.Count = 0 
                then
                    if partial_res.Count = sel //&& sumTo1(partial_res) 
                    then  Set.singleton partial_res
                    else Set.empty
                else
                    let l_array = Set.to_array l
                    seq {
                    for i in [0.. l_array.Length - 1]
                        -> 
                        let (s,interv,s1) = l_array.[i]
                        
                        //costruisco l'insieme di transizioni a cui levo quella scelta
                        let temp2_l, temp2_2 = 
                            let out =  ref Set.empty
                            let out2 =  ref Set.empty
                            for j=0 to l_array.Length - 2 do
                                if not (l_array.[j] = l_array.[i]) then  out := Set.add l_array.[j] !out; out2 := Set.add l_array.[j] !out2;
                            done
                            !out , !out2
                            
        
                        //print_endline((tr = (s,interv,s1)).ToString() + " " + (l).Contains(tr).ToString() + " " + (!temp_l).Count.ToString() + "  " + l.Count.ToString())
                        //print_endline((cut(Set.of_array temp2_l, acc + interv.low) = Set.of_array temp2_l).ToString())
                        
                        let cut1 = cutMaxs(cutMins(temp2_l, acc + interv.low), acc + interv.low)
                        let cut2 = cutMaxs(cutMins(temp2_2, acc + interv.up), acc + interv.up)
                        
                        Set.union
                            (pippo(cut1, acc + interv.low ,Set.add (new mdp_alternative(interv.low, s1)) partial_res,sel ))
                            (pippo(cut2, acc + interv.up, Set.add (new mdp_alternative(interv.up, s1)) partial_res,sel ))
                    //done
                    } |> Set.unionMany
                    


        let timer = System.Diagnostics.Stopwatch.StartNew ()
        let result_mdp : mdp_transition Set ref                      
                                             = ref Set.empty
        for s in input.S do
            let s_i = input.index s
            let current_transitions = input.get_i_transitions_from(s_i)
            
            let currrent_alternatives = 
                pippo(current_transitions, 0.0, Set.empty, current_transitions.Count)
                //!result
            for el in currrent_alternatives do
                let myNewEntry = new mdp_transition
                                        (
                                            "" , //label es> min Ri - di s1 -> s2
                                            s_i,
                                            el 
                                        )
                result_mdp := Set.add (myNewEntry) (!result_mdp)
            done
        done
        printfn "Time taken serial = %O" timer.Elapsed
        
        
  (*  ///PARALLEL TRY 1
        let inline task s = async { 
            do (
                let s_i = input.index s
                let current_transitions = input.get_i_transitions_from(s_i)
                
                let currrent_alternatives = 
                    pippo(current_transitions, 0.0, Set.empty, current_transitions.Count)

                for el in currrent_alternatives do
                    let myNewEntry = new mdp_transition
                                            (
                                                "" , //label es> min Ri - di s1 -> s2
                                                s_i,
                                                el 
                                            )
                    result_mdp := Set.add (myNewEntry) (!result_mdp)
                done
                
               
                )
                
            }
        
        let inline go() = seq {
                    for s in input.S
                        -> 
                    task s  
                    }
        
        match System.Environment.ProcessorCount with
            | 1 ->  Async.Run (Async.Parallel (go())) |> ignore 
            | n ->  Async.Run (Async.Parallel (go())) |> ignore 

        //Async.Run (Async.Parallel (go())) |> ignore 
*)

(*
     //parallell try 2
        let map_parallel (func : 'a -> 'b) (items : 'a[]) =
            let results = Array.zero_create (items.Length)
            let count = ref items.Length
            use notify = new System.Threading.AutoResetEvent(false)

            items
            |> Array.iteri (
                fun i item ->
                System.Threading.ThreadPool.QueueUserWorkItem (
                    fun _ ->
                    let res = func item
                    results.[i] <- res
                    //let c = System.Threading.Interlocked.Decrement count
                    if (System.Threading.Interlocked.Decrement count) = 0 then notify.Set() |> ignore
                ) |> ignore
            )
            notify.WaitOne() |> ignore
            results
    
        let myFunc s = 
            let s_i = input.index s
            let current_transitions = input.get_i_transitions_from(s_i)
            
            let currrent_alternatives = 
                pippo(current_transitions, 0.0, Set.empty, current_transitions.Count)
                //!result
                
            seq {
                for el in currrent_alternatives  ->
                
                Set.singleton( new mdp_transition
                                        (
                                            "" , //label es> min Ri - di s1 -> s2
                                            s_i,
                                            el 
                                        )
                            )
                  } |> Set.union_all
        let timer = System.Diagnostics.Stopwatch.StartNew ()          
        let result = (Set.of_array ( map_parallel myFunc (Set.to_array (input.S))))
        let result_mdp = 
            ref 
                (
                let res = ref Set.empty
                for el in result do
                    for el2 in el do
                        res := Set.add (el2) !res
                    done
                done
                
                !res
                )
            
        
        printfn "Time taken parallel = %O" timer.Elapsed
        
        let eq = Seq.for_all2 (fun a b -> if a = b 
                                                then true 
                                                else print_endline(a.ToString() + " " + b.ToString())
                                                     false
                                                     ) (!result_mdp2) (!result_mdp1)
        
        print_any(eq)
         *)
        let S = [0..input.S.Count-1] |> Set.of_seq 
        //let b = Set.to_array ( Set.of_seq (!result_mdp))
        //let b1 = Set.to_array ( Set.of_seq (!result_mdp1))
        //print_any(b=b1)
        new mdp(!result_mdp, S, input.S, input.s0, input.maximum_values_state)
        
        
    member inline x.add(t:mdp_transition) = 
        let involved_states =
            let mutable out= Set.singleton t.s
            for mdp_tr in t.alternatives do out <- out.Add(mdp_tr.s1) done
            out
        x.transitions <- x.transitions.Add(t)
        x.S <- Set.union x.S involved_states
        
    member inline x.remove(t:mdp_transition) = 
        x.transitions <- x.transitions.Remove(t)
        
    override x.ToString() =
        //costruzione stringa
        let inline buildString(input: mdp_transition Set, state_number) =
            let indice_stato_iniziale = Array.findIndex (fun el -> el = x.s0) (x.multisetsS |> Set.to_array)
            let out = new StringBuilder("mdp \n \n module M \n state : [0.."+ (state_number - 1).ToString() + "] init "+indice_stato_iniziale.ToString() + ";\n")
            for tr in  input do
                let label, s1, alternatives = tr.label, tr.s, tr.alternatives
                out.Append("[" + label + "] state="+s1.ToString() + " -> ") |> ignore
                for mdp_tr in alternatives do
                    let probability_string = mdp_tr.prob.ToString() 
                    out.Append(" " + probability_string.Replace(",",".") + " : (state'=" + mdp_tr.s1.ToString() + ") +" ) |> ignore
                done   
                (if alternatives.Count  > 0 then out.Remove(out.Length - 1, 1 )|> ignore)  
                out.Append(";\n") |> ignore
            done
            out.Append("\n endmodule").ToString()
        
        buildString(x.transitions, x.S.Count)
    
    
    member  x.ToPrismModel(symbols: string SymbolTable) = 
        //costruzione stringa
            let states:string array = x.PRISMStates(symbols)
            let states_as_arrival:string array = x.PRISMStatesAsArrival(symbols)
            let transitions: ResizeArray<Set<string * int>> array = x.PRISMTransitions()
            
            let out = new StringBuilder("mdp \n \n");
            
            for i = 0 to 5 do 
                out.AppendLine("const int k" + i.ToString() + ";") |> ignore 
            done 
            
            out.Append("\n module M \n") |> ignore   // state : [0.."+ (state_number - 1).ToString() + "] init "+indice_stato_iniziale.ToString() + ";\n")

            let findMaxOnVar(i) =
                Set.fold (fun (acc:int) (el:multiset) -> if el.ar.[i] > acc then el.ar.[i] else acc) 0 x.multisetsS 
                
            
            let maximum_values =
                let max_state = (x.multisetsS |> Set.to_array ).[ x.maximum_values_state]
                let mutable out = Array.zeroCreate max_state.size
                for i=0 to max_state.size - 1 do
                    out.[i] <- findMaxOnVar(i) + 1//max_state.ar.[i]
                done
                out
            
            
            let inline variables(st: string SymbolTable) =
                let mutable output = new ResizeArray<string>();
                for keyVal in st.table do 
                    output.Add(keyVal.Value.ToString()) 
                done
                output.ToArray()
            
            let vars = variables(symbols)
            
            for i=0 to vars.Length - 1 do
                out.Append(vars.[i] + " : [0.." + maximum_values.[i].ToString() + "] init " + x.s0.ar.[i].ToString() + "; \n") |> ignore
                ()
            done
                               
            for i=0 to  states.Length - 1 do
                for  ndt_alternatives in transitions.[i] do
                    out.Append("[ ]" +  states.[i] + " -> ") |> ignore
                    for (p_strin, state_index) in ndt_alternatives do
                        out.Append(" " + p_strin + " : " + states_as_arrival.[state_index] + " +" ) |> ignore
                    done   
                    (if transitions.[i].Count  > 0 then out.Remove(out.Length - 1, 1 )|> ignore)  
                    out.Append(";\n") |> ignore
                done
            done
            out.Append("\n endmodule").ToString()
      
    member inline private x.ToPRISMStates(symbols : string SymbolTable) =
        let inline ToPRISMString(st: string SymbolTable) =
            let mutable output = new StringBuilder("(")
            for keyVal in st.table do 
                output.Append(keyVal.Value.ToString() + "," ) |> ignore
                ()
            (if st.table.Count  > 0 then output.Remove(output.Length - 1, 1 )|> ignore)   
            output.Append(")") |> ignore
            output.ToString()
            
        let out_states = new StringBuilder()
        //(v1 , v2 ..)
        out_states.AppendLine(ToPRISMString(symbols)) |> ignore
        //id:(v1=val, v2=val)
        //for s in x.multisetsS do
        let indexedS = x.S |> Set.to_array //|> Array.rev
        let indexedMS = x.multisetsS |> Set.to_array //|> Array.rev
        for i=0 to indexedS.Length - 1 do
                //le transizioni in uscita corrispondenti 
            let current_trans = x.transitions |> Set.filter (fun (tr) -> tr.s = indexedS.[i] )
            let current_state_index = indexedS.[i]
            let current_state = indexedMS.[current_state_index]       
            out_states.AppendLine((indexedS.[i]).ToString() + ":(" +  current_state.ToPRISMString() + ")") |> ignore 
        done
        out_states.ToString()
   
   
    member private x.PRISMStates(symbols : string SymbolTable) =
        let inline variables(st: string SymbolTable) =
            let mutable output = new ResizeArray<string>();
            for keyVal in st.table do 
                output.Add(keyVal.Value.ToString()) 
            done
            output.ToArray()
            
        let vars = variables(symbols)
        let out_states = Array.create x.multisetsS.Count ""

        let indexedS = x.S |> Set.to_array //|> Array.rev
        let indexedMS = x.multisetsS |> Set.to_array //|> Array.rev
        for i=0 to indexedS.Length - 1 do
                //le transizioni in uscita corrispondenti 
            let current_trans = x.transitions |> Set.filter (fun (tr) -> tr.s = indexedS.[i] )
            let current_state_index = indexedS.[i]
            let current_state = indexedMS.[current_state_index]
            let current_state_string = current_state.ToPRISMString2(vars)
            out_states.[(indexedS.[i])] <-  current_state_string
        done
        
        out_states   
        
    member private x.PRISMStatesAsArrival(symbols : string SymbolTable) =
        let inline variables(st: string SymbolTable) =
            let mutable output = new ResizeArray<string>();
            for keyVal in st.table do 
                output.Add(keyVal.Value.ToString()+"'") 
            done
            output.ToArray()
            
        let vars = variables(symbols)
        let out_states = Array.create x.multisetsS.Count ""

        let indexedS = x.S |> Set.to_array //|> Array.rev
        let indexedMS = x.multisetsS |> Set.to_array //|> Array.rev
        for i=0 to indexedS.Length - 1 do
                //le transizioni in uscita corrispondenti 
            let current_trans = x.transitions |> Set.filter (fun (tr) -> tr.s = indexedS.[i] )
            let current_state_index = indexedS.[i]
            let current_state = indexedMS.[current_state_index]
            let current_state_string = current_state.ToPRISMString2(vars)
            out_states.[(indexedS.[i])] <-  current_state_string
        done
        
        out_states   
    
    member inline private x.ToPRISMTransitions(symbols : string SymbolTable) =
        let inline getTransitionsInformation() =
            let nr_states = x.S.Count
            let nr_transitions_ndt = x.transitions.Count
            let nr_transitions =
                Set.fold (fun acc (trSet:mdp_transition) -> trSet.alternatives.Count + acc) 0 x.transitions
            nr_states.ToString() + " " + nr_transitions_ndt.ToString() + " " + nr_transitions.ToString()
            
        //nr_stati  nr_transizioni_ndt  nr_transizioni
        let out_transitions = new StringBuilder(getTransitionsInformation()+"\n")
        //s action_index    s1  prob
        
        //per ogni stato di partenza
        let indexedS = x.S |> Set.to_array 
        for i=0 to indexedS.Length - 1 do
                //le transizioni in uscita corrispondenti 
            let current_trans = x.transitions |> Set.filter (fun (tr) -> tr.s = i )
            let j = ref 0
            for tr in current_trans do
                let label,s,alternatives = tr.label, tr.s, tr.alternatives
                for alternative in alternatives do
                    let prob, s1 = alternative.prob, alternative.s1
                    let probstring = prob.ToString()
                    out_transitions.AppendLine(s.ToString() + " " + (!j).ToString() (*AZIONE*) + " " + s1.ToString() + " " + probstring.Replace(',','.') + " " ) |> ignore
                    ()
                done
                j := !j + 1
            done
        out_transitions.ToString()
    
    member private x.PRISMTransitions() : ResizeArray<Set<string * int>> array =
        let out_transitions = Array.zeroCreate x.multisetsS.Count 
     
        //per ogni stato di partenza
        let indexedS = x.S |> Set.to_array 
        for i=0 to indexedS.Length - 1 do
            
                //le transizioni in uscita corrispondenti 
            let current_trans = x.transitions |> Set.filter (fun (tr) -> tr.s = i )
            let j = ref 0
            out_transitions.[i] <- new ResizeArray<_>()
            for tr in current_trans do
                let label,s,alternatives = tr.label, tr.s, tr.alternatives
                let mutable alternatives_set = Set.Empty
                for alternative in alternatives do
                
                    let prob, s1 = alternative.prob, alternative.s1
                    let probstring = prob.ToString()
                    alternatives_set <- alternatives_set.Add((probstring.Replace(',','.'), s1))
                    //out_transitions.AppendLine(s.ToString() + " " + (!j).ToString() (*AZIONE*) + " " + s1.ToString() + " " + probstring.Replace(',','.') + " " ) |> ignore
                    done
                out_transitions.[i].Add(alternatives_set)
                done
            j := !j + 1
            
        done
        out_transitions
        
    member inline private x.ToPRISMInitial(symbols : string SymbolTable) =
        //v1=val, v2=val, v3=val
        x.s0.ToPRISMInitString(symbols)
    
    member inline x.ToPRISMTransitionsAndStates(symbols : string SymbolTable) =
        x.ToPRISMStates(symbols) , x.ToPRISMTransitions(symbols), x.ToPRISMInitial(symbols)
    
    member inline x.ToGLEEGraph() : Microsoft.Glee.Drawing.Graph= 
            let g = new Microsoft.Glee.Drawing.Graph("graph");
            g.Cluster <- true;
            g.Directed <- true;
            g.GraphAttr.NodeAttr.Padding <- 50.0;
            g.GraphAttr.LayerDirection<- Microsoft.Glee.Drawing.LayerDirection.LR;
            g.GraphAttr.NodeAttr.LineWidth <- 2
            g.GraphAttr.EdgeAttr.LineWidth <- 2
            //g.GraphAttr.NodeAttr.FontName <- "Lucida-Sans"
           

            let sigma = ['a' .. 'z']  |>  Seq.to_array 
            let indexedStates = x.S |> Set.to_array //|> Array.rev
            let indexedMsStates = x.multisetsS |> Set.to_array //|> Array.rev
            //per ogni stato di partenza
            for i=0 to indexedStates.Length - 1 do
                if indexedMsStates.[i] = x.s0 then
                    let vv = g.AddNode("s"+i.ToString()^" : "^indexedMsStates.[i].ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.Red
                    vv.NodeAttribute.LineWidth <- 6
                if x.maximum_values_state > -1 &&indexedStates.[i] = x.maximum_values_state then
                    let vv = g.AddNode("s"+i.ToString()^" : "^indexedMsStates.[i].ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.LightGreen
                    vv.NodeAttribute.LineWidth <- 6
                //le transizioni in uscita corrispondenti 
                let current_trans = x.transitions |> Set.filter (fun (tr) -> tr.s = indexedStates.[i])
                let j = ref 0
                for tr in current_trans do
                    
                    let label,s,alternatives = tr.label, tr.s, tr.alternatives
                    let current_action = sigma.[!j]
                    //dallo stato all'azione
                    let vv = g.AddNode(current_action.ToString()+"-"+"s"+s.ToString())
                    vv.NodeAttribute.Color <- Microsoft.Glee.Drawing.Color.Blue
                    let v = g.AddEdge("s"+s.ToString()^" : "^indexedMsStates.[s].ToString(), label ,current_action.ToString()+"-"+"s"+s.ToString()) //|> ignore // directed inverted
                    v.EdgeAttr.Color <- Microsoft.Glee.Drawing.Color.Blue
                    

                    for mdpt_tr in alternatives do
                        let fraction_string = Model.fractionString(mdpt_tr.prob)
                        let v = g.AddEdge(current_action.ToString()+"-"+"s"+s.ToString(), fraction_string ,"s"+ mdpt_tr.s1.ToString()^" : "^ indexedMsStates.[ mdpt_tr.s1].ToString()) //|> ignore // directed inverted
                        ()
                        //v.EdgeAttr.Color <- Color.Gray
                    done
                    j := !j + 1
                    ()
                done
                
                ()
            done
            g
    
    end

    
 and Model =
    class
    val symbols : string SymbolTable
    val rules : ab_rw_rule array
    val term : multiset
    val properties : property array
    val mutable state_space_limit : int
    new(rules,term,symbols,properties,limit) = {symbols=symbols; rules=rules; term=term; properties=properties; state_space_limit=limit;}
    new(rules,term,symbols,properties) = {symbols=symbols; rules=rules; term=term; properties=properties; state_space_limit=Model.default_state_space_limit;}
    static member default_state_space_limit = 10000;
    static member DEBUG = true;
    override x.ToString() =
            let rulesString =
                let out = new StringBuilder()
                for r in x.rules do
                    out.Append(r.ToString(x.symbols)) |> ignore
                done
                out.ToString()
            "species \n" + x.symbols.ToString() + "\n rules\n" + rulesString + "\n term\n" + x.term.ToString(x.symbols)
    member inline x.computeBeta(n:int, s1:multiset) =
        let mutable out = 1.0
        for i in [0..(s1.size - 1)] 
            do
            let a = s1.ar.[i]
            let b = x.rules.[n].l.ar.[i]
            out <- out * MyMath.Binomial (s1.ar.[i]) (x.rules.[n].l.ar.[i])
            done
        let outL = out * x.rules.[n].ki.low
        let outH = out * x.rules.[n].ki.up
        new interval(outL, outH)
        
    member inline x.computeS2(n:int, s1:multiset) =
        let a = x.rules.[n].r 
        let b = multiset.sub (x.rules.[n].l)  s1
        multiset.add2(a, b)
        
    member (*inline*) x.toLTS() = /////////////////////////////////////////DA RIVEDERE
        let mutable active_states = Set.singleton x.term;
        let mutable viewed_states = Set.empty;
        let mutable S = Set.empty;
        let s0 = x.term;
        let rules = x.rules;
        let tr = new ab_LTS_transition_relation();
        
        let inline probabilita_considerevole(b) =
            not (b = new interval(0.0, 0.0))
            //let epsilon_min, epsilon_max = new interval(0.0000000001, 0.0000000010)
            //b <= new interval(epsilon_min,epsilon_max)
        
        ///////////
        let mutable view_states = Set.Empty
        
        while (S.Count <= x.state_space_limit) && (active_states.Count > 0) 
            do 
            let mutable to_remove_from_active = Set.empty
            let mutable to_add_to_active = Set.empty
            
            for s1 in active_states 
                do
                //add s1 to S
                //let a = S.Count
                S <- S.Add(s1)  
                (*if true //a <> S.Count 
                                    then System.Console.WriteLine("I add" + s1.ToString() + ": " +  s1.GetHashCode().ToString() +   " : " +  a.ToString()+ " -> " + S.Count.ToString())
                                         for s in S do System.Console.WriteLine(s.ToString() + " : " + s.GetHashCode().ToString());
                                         System.Console.WriteLine("")
                                    else ()*)
                //remove s1 from active states
                to_remove_from_active <- Set.add s1 to_remove_from_active
                //viewed_states <- viewed_states.Add(s1)// viewed_states
                for n in [0..(x.rules.Length - 1)]
                    do
                    if multiset.submultiset x.rules.[n].l s1
                        then
                            let b = x.computeBeta(n,s1) 
                            if probabilita_considerevole(b) then    //continuo solo se l'intervallo è diverso da [0,0]
                                let s2 = x.computeS2(n,s1)
                                //let a = S.Count
                                S <- S.Add(s2)  
                                
                                (*if true //a <> S.Count  
                                    then System.Console.WriteLine("I add" + s2.ToString() + ": " + s2.GetHashCode().ToString() +   " : " +  a.ToString()+ " -> " + S.Count.ToString())
                                         for s in S do System.Console.WriteLine(s.ToString() + " : " + s.GetHashCode().ToString());
                                         System.Console.WriteLine("")
                                    else ()*)
                                if not (view_states.Contains(s2)) then to_add_to_active  <- to_add_to_active.Add(s2 ) 
                                //System.Console.WriteLine("attivo" + s2.ToString())
                                tr.add(s1,n,b,s2)
                    done
                view_states <- view_states.Add(s1)
                //active_states <- Set.union active_states to_add_to_active 
                //active_states <- Set.diff active_states to_remove_from_active
                done
            active_states <-  active_states + to_add_to_active 
            active_states <- active_states - to_remove_from_active
            done
            
        /////////// //quando ho esaurito lo spazio massimo per lo state space, continuo comunque ad aggiungere le transizioni dagli stati attivi a stati già visitati
        let maximum_values_state =
            if S.Count >= x.state_space_limit   
                then
                    
                    ///inoltre aggiungo uno stato speciale in cui vado da tutti gli stati limite ed in cui ciclo
                    ///questo stato ha come valore un valore superiore per ogni reagente di ogni stato nello spazio degli stati
                    let maximum_values_state =
                        let state_size = x.term.size
                        let stato = new multiset(state_size)
                        let accumula_il_massimo (acc_stato:multiset) (s:multiset) =  
                                                        for i=0 to state_size - 1 do 
                                                            if s.ar.[i] > acc_stato.ar.[i] then acc_stato.ar.[i] <- s.ar.[i];
                                                            done
                                                        acc_stato  
                        let massimo = Set.fold accumula_il_massimo stato S 
                        for i=0 to state_size - 1 do 
                            massimo.ar.[i] <- massimo.ar.[i] + 1
                        done
                        massimo
                    
//                    let minimum_values_state =
//                        let state_size = x.term.size
//                        let stato = new multiset(state_size)
//                        let accumula_il_minimo (acc_stato:multiset) (s:multiset) =  
//                                                        for i=0 to state_size - 1 do 
//                                                            if s.ar.[i] < acc_stato.ar.[i] then acc_stato.ar.[i] <- s.ar.[i];
//                                                            done
//                                                        acc_stato  
//                        let minimo = Set.fold accumula_il_minimo stato S 
//                        for i=0 to state_size - 1 do 
//                            minimo.ar.[i] <- max( minimo.ar.[i] - 1, 0)
//                        done
//                        minimo
                    
                    
                    
                    S <- S.Add(maximum_values_state)
                    tr.add(maximum_values_state, 0, new interval(1.0, 1.0), maximum_values_state)
                    
                    if (active_states.Count > 0)
                        then
                        for s1 in active_states 
                            do
                            for n in [0..(x.rules.Length - 1)]
                                do
                                if multiset.submultiset x.rules.[n].l s1
                                    then
                                        //aggiungo le transizioni verso stati che già conosco
                                        let b = x.computeBeta(n,s1) 
                                        if probabilita_considerevole(b) then    //continuo solo se l'intervallo è diverso da [0,0]
                                            let s2 = x.computeS2(n,s1)
                                            if S.Contains(s2)
                                                then    tr.add(s1,n,b,s2)
                                                else    tr.add(s1,n,b,maximum_values_state)
                                done
                            done
                    maximum_values_state
                else new multiset(0)
                ///////////
        new ab_LTS(S,s0,rules,tr,maximum_values_state);
    
    member x.toIMC() =
        x.toIMC(x.toLTS())
    
    member x.toIMC(lts) =
        let ptr = new ab_MC_transition_relation()
        //let a = lts.S |> Set.of_seq
        //System.Console.WriteLine(a.Count.ToString())
        for s1 in lts.S  do
            let exit_rate_interval = lts.getExitRate(s1)
            match (exit_rate_interval.low,exit_rate_interval.up) with
                | (n,0.0) -> 
                    for s2 in lts.getNext(s1) do
                        ptr.add(s1,new interval(0.0, 0.0), s2)
                    done
                    ptr.add(s1,new interval(1.0, 1.0), s1)
                | _ -> 
                    if exit_rate_interval.low = 0.0
                        then
                            ptr.add(s1,new interval(0.0, 1.0), s1)
                    for s2 in lts.getNext(s1) do
                        let rate = lts.getRate(s1,s2)
                        let low_tr_p_s1_s2 = if rate.low = 0.0 then 0.0 else (rate.low) / (exit_rate_interval.up - rate.up + rate.low)
                        let high_tr_p_s1_s2 = if rate.up = 0.0 then 0.0 else (rate.up) / (exit_rate_interval.low - rate.low + rate.up)
                        ptr.add(s1,new interval(low_tr_p_s1_s2, high_tr_p_s1_s2), s2)
                    done        
        done
        new ab_markov_chain(lts.S, lts.s0, x.rules, ptr, lts.maximum_values_state)
    
    static member fractionString(value:float) = 
        let mutable out = value.ToString()
        try 
            let rat = new Rational(value)
            out <- rat.ToString()
        with | e -> ()
        out
        
            
      
    static member cutIMC(input:ab_markov_chain) =
        let myPtr = new ab_MC_transition_relation()
        //per ogni tr=(s1, s2, i_min, i_max)
        let tr = input.getPtr().tr
        for s1 in input.S do
            let current_trans = input.getPtr().getTransitionsFrom(s1)
            for t in current_trans do
                let a,b,c = t
                let k =
                    Set.filter (fun el -> let s1,b,s2 = el in not (s1 = s2)) current_trans |>  ///rivedere qui
                    Set.fold (fun acc el -> let s1,b,s2 = el in acc + b.up) 0.0 
                    
                let b1 = new interval(max (b.low) (1.0 - k), b.up)
                let t1 = (a,b1,c)
                myPtr.tr <- myPtr.tr.Add(a,b1,c)
            done
        done
        //creo tr1=(s1,s2,i_min1,i_max)
        //con i_min1 = max(i_min, Somm per tr(s1,s3) con s2!=s3 di tr.i_max)
        new ab_markov_chain(input, myPtr )
    
    member inline x.toDelimitedIMC() =
        Model.cutIMC(x.toIMC(x.toLTS()))
    
    
    
    static member inline toMDPProperties(input:ab_markov_chain, properties : property array) =
        let indexedStateSpace = (*Array.rev*) (Set.to_array input.S)
        let mapPropertyToSatisfiedStates =
            fun (p:property) 
                ->
                let predicate(a) =
                    match p.relation with
                        | rel.equal -> a = p.value
                        | rel.great -> a > p.value
                        | rel.geq -> a >= p.value
                        | rel.leq -> a <= p.value
                        | rel.less -> a < p.value
                        | _ -> failwith ""
                        
                Set.filter (fun (s:multiset) -> predicate(s.ar.[p.id]) ) input.S
                |>
                Set.map (fun s -> Array.findIndex (fun el -> el = s) indexedStateSpace)  
            
        let propertiesInTermsOfStates = Array.map mapPropertyToSatisfiedStates properties 
        
        let out = new StringBuilder()
        
        for i=0 to properties.Length - 1 do
            let current_p = properties.[i]
            let current_states = propertiesInTermsOfStates.[i]
            
            out.AppendLine("// " + current_p.decl_string) |> ignore
            
            out.Append("Pmax=? [ F" ) |> ignore
            match propertiesInTermsOfStates.[i].Count with
                | 0 -> out.Append(" false ")|>ignore
            
                | n ->  if n =  indexedStateSpace.Length then  out.Append(" true ")|>ignore
                        else
                            for s_index in propertiesInTermsOfStates.[i] do
                                out.Append("(state="+s_index.ToString()+") | " ) |> ignore 
                            done
                            (if propertiesInTermsOfStates.Length  > 0 then out.Remove(out.Length - 2, 2)|> ignore)  
            out.Append("]\n")|>ignore
            
            out.Append("Pmin=? [ F" ) |> ignore
            match propertiesInTermsOfStates.[i].Count with
                | 0 -> out.Append(" false ")|>ignore
                | n ->  if n =  indexedStateSpace.Length then  out.Append(" true ")|>ignore
                        else
                            for s_index in propertiesInTermsOfStates.[i] do
                                out.Append("(state="+s_index.ToString()+") | " ) |> ignore 
                            done
                            (if propertiesInTermsOfStates.Length  > 0 then out.Remove(out.Length - 2, 2)|> ignore)  
            out.Append("]\n\n")|>ignore
        done
        out.ToString()
        
    static member inline toMDPProperties_directly(input:ab_markov_chain, properties : property array) =
       
        let out = new StringBuilder()
        
        for i=0 to properties.Length - 1 do
            let current_p = properties.[i]
            out.AppendLine("// " + current_p.decl_string) |> ignore
           
            out.Append("Pmax=? [ F" ) |> ignore
            out.Append(" ( " + current_p.decl_string + " ) " ) |> ignore   
            out.Append("]\n")|>ignore
            
            out.Append("Pmin=? [ F" ) |> ignore
            out.Append(" ( " + current_p.decl_string + " ) " ) |> ignore   
            out.Append("]\n")|>ignore
            
        done
        out.ToString() 
    
   
    (*
    member x.toMDP() =
        Model.toMDP(x.toIMC(x.toLTS()))
    *)

    
  
    
    static member saveGraphImage(abMC:ab_markov_chain,outFileName:string) =
            let working_dir = System.IO.Directory.GetCurrentDirectory();
            let graph = abMC.ToGLEEGraph()
            let renderer = Microsoft.Glee.GraphViewerGdi.GraphRenderer(graph)
            renderer.CalculateLayout()
            let bmp = new System.Drawing.Bitmap((int graph.Width), (int graph.Height), System.Drawing.Imaging.PixelFormat.Format32bppArgb)
            renderer.Render(bmp)
            bmp.Save(outFileName)
            System.IO.Directory.SetCurrentDirectory(working_dir)
            ()
            
    static member inline saveLTSGraphImage(lts:ab_LTS,outFileName:string) =
            let working_dir = System.IO.Directory.GetCurrentDirectory();
            let graph = lts.ToGLEEGraph()
            let renderer = Microsoft.Glee.GraphViewerGdi.GraphRenderer(graph)
            renderer.CalculateLayout()
            let bmp = new System.Drawing.Bitmap((int graph.Width), (int graph.Height), System.Drawing.Imaging.PixelFormat.Format32bppArgb)
            renderer.Render(bmp)
            bmp.Save(outFileName)
            System.IO.Directory.SetCurrentDirectory(working_dir)
            ()
        
    static member inline saveMDPGraphImage(myMDP:mdp,outFileName:string) =
            let working_dir = System.IO.Directory.GetCurrentDirectory();
            let graph = myMDP.ToGLEEGraph()
            let renderer = Microsoft.Glee.GraphViewerGdi.GraphRenderer(graph)
            renderer.CalculateLayout()
            let bmp = new System.Drawing.Bitmap((int graph.Width), (int graph.Height), System.Drawing.Imaging.PixelFormat.Format32bppArgb)
            renderer.Render(bmp)
            bmp.Save(outFileName)
            System.IO.Directory.SetCurrentDirectory(working_dir)
            ()
     
    end