using System.Collections;
using System.Collections.Generic;
using System.Text;

using AMSR;

using System;



public class Parser {
	const int _EOF = 0;
	const int _Id = 1;
	const int _Number = 2;
	const int _RESULTDIC = 3;
	const int maxT = 23;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;
	
	public Scanner scanner;
	public Errors  errors;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;

public static List<double> ResultsList;



	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (string msg) {
		if (errDist >= minErrDist) errors.SemErr(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}


	bool WeakSeparator(int n, int syFol, int repFol) {
		int kind = la.kind;
		if (kind == n) {Get(); return true;}
		else if (StartOf(repFol)) {return false;}
		else {
			SynErr(n);
			while (!(set[syFol, kind] || set[repFol, kind] || set[0, kind])) {
				Get();
				kind = la.kind;
			}
			return StartOf(syFol);
		}
	}

	
	void PRISMRES() {
		PRISM();
	}

	void PRISM() {
		ResultsList = new List<double>();
		
		RESULT();
		if (la.kind == 3) {
			while (la.kind == 3) {
				RESULT();
			}
		}
	}

	void RESULT() {
		StringBuilder TEMP = new StringBuilder();
		Expect(3);
		if (la.kind == 4) {
			Get();
			Expect(5);
			Expect(6);
			Expect(7);
			Expect(8);
			Expect(9);
			Expect(10);
			Expect(2);
			Expect(11);
			Expect(12);
			Expect(13);
			Expect(14);
			Expect(15);
			Expect(16);
			Expect(6);
			Expect(17);
			Expect(18);
			Expect(19);
			Expect(20);
			Expect(21);
			Expect(22);
			Expect(11);
			ResultsList.Add(-1.0); 
		} else if (la.kind == 2) {
			rate(TEMP);
			double temp;
			double.TryParse(TEMP.ToString(), out temp); //.Replace('.',',')
			ResultsList.Add(temp);
			
		} else SynErr(24);
	}

	void rate(StringBuilder TEMP) {
		Expect(2);
		TEMP.Append(t.val); 
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		PRISMRES();

    Expect(0);
	}
	
	bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x}

	};
} // end Parser


public class Errors {
	public int count = 0;                                    // number of errors detected
	public System.IO.TextWriter errorStream = Console.Out;   // error messages go to this stream
  public string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
  
	public void SynErr (int line, int col, int n) {
		string s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "Id expected"; break;
			case 2: s = "Number expected"; break;
			case 3: s = "RESULTDIC expected"; break;
			case 4: s = "\"Error:\" expected"; break;
			case 5: s = "\"Iterative\" expected"; break;
			case 6: s = "\"method\" expected"; break;
			case 7: s = "\"did\" expected"; break;
			case 8: s = "\"not\" expected"; break;
			case 9: s = "\"converge\" expected"; break;
			case 10: s = "\"within\" expected"; break;
			case 11: s = "\"iterations.\" expected"; break;
			case 12: s = "\"Consider\" expected"; break;
			case 13: s = "\"using\" expected"; break;
			case 14: s = "\"a\" expected"; break;
			case 15: s = "\"different\" expected"; break;
			case 16: s = "\"numerical\" expected"; break;
			case 17: s = "\"or\" expected"; break;
			case 18: s = "\"increasing\" expected"; break;
			case 19: s = "\"the\" expected"; break;
			case 20: s = "\"maximum\" expected"; break;
			case 21: s = "\"number\" expected"; break;
			case 22: s = "\"of\" expected"; break;
			case 23: s = "??? expected"; break;
			case 24: s = "invalid RESULT"; break;

			default: s = "error " + n; break;
		}
		errorStream.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void SemErr (int line, int col, string s) {
		errorStream.WriteLine(errMsgFormat, line, col, s);
		count++;
	}
	
	public void SemErr (string s) {
		errorStream.WriteLine(s);
		count++;
	}
	
	public void Warning (int line, int col, string s) {
		errorStream.WriteLine(errMsgFormat, line, col, s);
	}
	
	public void Warning(string s) {
		errorStream.WriteLine(s);
	}
} // Errors


public class FatalError: Exception {
	public FatalError(string m): base(m) {}
}

