﻿//-----------------------------------------------------------------------------

//          file: $Archive:$

//    created by: Johannes Weigend

// creation date: 14/11/2000

//    changed by: $Author: $

//   change date: $Date:  $

//   description: Rational Numbers a/b - C# Version

//       version: $Revision: 1.5 $

//     copyright: (c) weigend software engineering 2001

//-----------------------------------------------------------------------------

//

// JAVA Style Namespace

//

namespace myMath 

{


	using System;

	using System.Collections;


	/// <summary>

	///    Rational Numbers a/b

	/// </summary>

	public struct Rational : ICloneable, IComparable

	{

		/// <summary>

		///    Constructor take nominator and denominator

		///    <exception>Division by Zero</exception>

		/// </summary>

		public Rational(long nominator, long denominator) 

		{

			if (denominator == 0) throw new Exception("Division by Zero");


			this.nominator = nominator;

			this.denominator = denominator;

			norm();

		}

		

		/// <summary>

		///    Delegating Constructor takes double

		///    <exception>Division by Zero</exception>

		/// </summary>

		public Rational(double d) 

			: this((long) (d * 1000.0), (long) 1000) {}

		

		/// <summary>

		///    Property Nominator

		/// </summary>

		public long nom 

		{

			get { return nominator; }

		}

		

		/// <summary>

		///    Property Denominator

		/// </summary>

		public long denom 

		{

			get { return denominator; }

		}

		

		/// <summary>

		///    Indexer

		/// </summary>

		public long this[int index] 

		{

			get 

			{

				if (index == 0) return nominator;

				else if (index == 1) return denominator;

				else throw new Exception("Index out of Range");

			}

		}

		

		public static Rational operator * (Rational r1, Rational r2) 

		{

			return new Rational(r1.nominator * r2.nominator, r1.denominator * r2.denominator);

		}

		

		public static Rational operator + (Rational r1, Rational r2) 

		{

			return new Rational(r1.nominator * r2.denominator +

				r2.nominator * r1.denominator,

				r1.denominator * r2.denominator);

		}

		

		public static Rational operator - (Rational r1, Rational r2) 

		{

			return new Rational(r1.nominator * r2.denominator -

				r2.nominator * r1.denominator,

				r1.denominator * r2.denominator);

		}


		public static bool operator == (Rational r1, Rational r2)

		{

			return r1.Equals(r2); 

		}


		public static bool operator != (Rational r1, Rational r2)

		{

			return !(r1 == r2);

		}

		/// <summary>

		/// Sorting Support (IComparable)

		/// </summary>

		public int CompareTo(object o)

		{

			Rational r = ((Rational) this) - ((Rational) o);

			return (int) (r.nominator / r.denominator);

		}


		/// <summary>

		/// Hashtable Support

		/// </summary>

		public override int GetHashCode()

		{

			return nominator.GetHashCode() + denominator.GetHashCode();

		}

		

		public override bool Equals(object o)

		{

			if (!(o is Rational)) return false;

			Rational r = (Rational) o;

			return nominator == r.nominator && denominator == r.denominator;

		}

		

		/// <summary>

		/// implicit type conversion to double

		/// </summary>

		public static implicit operator double(Rational r)

		{

			return (double) r.nominator / (double) r.denominator;

		}

		

		/// <summary>

		/// implicit type conversion from Double

		/// </summary>

		public static implicit operator Rational(double d)

		{

			return new Rational(d);

		}

		

		/// <summary>

		/// object.ToString() Support

		/// </summary>

		public override string ToString()

		{

            if (denominator == (long) 1) return (String.Format("{0}", nominator));
			return(String.Format("{0}/{1}", nominator, denominator));

		}

		

		/// <summary>

		/// ICloneable-Support

		/// </summary>

		public object Clone()

		{

			return base.MemberwiseClone();

		}

		

		/// <summary>

		/// Private Helper

		/// </summary>

		private void norm()

		{

			long t = gcd(nominator, denominator);

			if (t != 1)

			{	

				nominator /= t;

				denominator /= t;

			}

		}

		

		/// <summary>

		/// Calculation of the greatest common devisor

		/// </summary>

		private static long gcd(long a, long b)

		{

			if (a < b) return gcd(b, a);

			

			long c = 1;

			while (c != 0)

			{

				//if (b != 0) {c = a % b;} PATCH!!!
                c = a % b;

				a = b;

				b = c;

			}			

			return a;

		}


		private long nominator;

		private long denominator;

	}

}

	


